ad_page_contract {
    Main view page for one task.

    @author jader@bread.com
    @creation-date 2003-07-28
    @cvs-id $Id: task-one.tcl,v 1.22.2.9 2004/03/12 00:20:24 jader Exp $

    @return task_term Term to use for Task
    @return task_term_lower Term to use for task
    @return assignee_term Term to use for assignee
    @return watcher_term Term to use for watcher
    @return dependency multirow that stores dependency information
    @return dependency2 multirow that stores dependency information for tasks that have dependencies on this particular task

    @param task_id item_id for the task
    @param project_item_id the item_id for the project. Used for navigational links
    @param project_id the revision_id for the project. Used for navigational links
    @param context_bar value for context bar creation
    @param orderby_revisions specifies how the revisions table will be sorted
    @param orderby_dependency specifies how the dependencies will be sorted
    @param orderby_dependency2 specifies how the dependencies will be sorted (for tasks that have dependencies on this task)
} {
    task_id:integer,optional
    task_revision_id:integer,optional
    orderby_revisions:optional
    orderby_dependency:optional
    orderby_dependency2:optional
} -properties {
    notification_chunk:onevalue
    task_info:onerow
    project_item_id:onevalue
    project_id:onevalue
    context:onevalue
    write_p:onevalue
    create_p:onevalue
    revisions:multirow
    dependency:multirow
    dependency2:multirow
    people:multirow
    task_term:onevalue
    task_term_lower:onevalue
    assignee_term:onevalue
    watcher_term:onevalue
    comments:onevalue
    comments_link:onevalue
    use_uncertain_completion_times_p:onevalue
} -validate {
    task_id_exists {
        if {![info exists task_id]} {
            set task_id [pm::task::get_item_id \
                             -task_id $task_revision_id]
            if {[string equal $task_id -1]} {
                ad_complain
            }
        }
    }
    revision_id_exists {
        if {![info exists task_revision_id]} {
            set task_revision_id [pm::task::get_revision_id \
                                      -task_item_id $task_id]
            if {[string equal $task_revision_id -1]} {
                ad_complain
            }
        }
    }
} -errors {
    task_id_exists {That task does not exist}
    revision_id_exists {That task does not exist}
}


# --------------------------------------------------------------- #

# terminology
set task_term       [parameter::get -parameter "TaskName" -default "Task"]
set task_term_lower [parameter::get -parameter "taskname" -default "task"]
set assignee_term   [parameter::get -parameter "AssigneeName" -default "Assignee"]
set watcher_term    [parameter::get -parameter "WatcherName" -default "Watcher"]
set project_term    [parameter::get -parameter "ProjectName" -default "Project"]
set use_uncertain_completion_times_p [parameter::get -parameter "UseUncertainCompletionTimesP" -default "1"]

db_1row get_project_ids { }

set context [list "one?project_item_id=$project_item_id $project_term" "View"]


# the unique identifier for this package
set package_id [ad_conn package_id]
set user_id    [ad_maybe_redirect_for_registration]

set comments [general_comments_get_comments -print_content_p 1 -print_attachments_p 1 $task_id "[ad_conn url]?task_id=$task_id"]

set comments_link [general_comments_create_link -object_name pm_task -link_text "Add a comment" -context_id $package_id $task_id "[ad_conn url]?task_id=$task_id"]



# permissions
permission::require_permission -party_id $user_id -object_id $package_id -privilege read

set write_p  [permission::permission_p -object_id $package_id -privilege write] 
set create_p [permission::permission_p -object_id $package_id -privilege create]



# Task info ----------------------------------------------------------

db_1row task_query { } -column_array task_info
set task_info(description) [ad_text_to_html -- $task_info(description)]

if {[exists_and_not_null task_info(earliest_start_j)]} {
    set task_info(slack_time) [pm::task::slack_time \
                                   -earliest_start_j $task_info(earliest_start_j) \
                                   -today_j $task_info(today_j) \
                                   -latest_start_j $task_info(latest_start_j)]
}

set logger_project [pm::project::get_logger_project \
                        -project_item_id $project_item_id]
set logger_variable_id [logger::variable::get_default_variable_id]

set log_url "[ad_conn package_url]log?project_id=$logger_project&pm_project_id=$project_item_id&pm_task_id=$task_id"

set log_note "Task logged time listed below"

# ------------------
# Notifications info
# ------------------
set notification_chunk [notification::display::request_widget \
                            -type pm_task_notif \
                            -object_id $task_id \
                            -pretty_name "$task_info(task_title)" \
                            -url "[ad_conn url]?[ad_conn query]" \
                           ]


# Task Revisions, using list-builder ---------------------------------

template::list::create \
    -name revisions \
    -multirow revisions \
    -key revision_id \
    -elements {
        revision_id {
            label "Subject"
            display_col task_title
            link_url_col item_url
            link_html { title "View this revision" }
            display_template {<if @revisions.live_revision@ eq @revisions.revision_id@><B>@revisions.task_title@</B></if><else>@revisions.task_title@</else>}
        }
        description {
            label "Description"
        }
        percent_complete {
            label "Status"
            display_template "@revisions.percent_complete@\%"
        }
        actual_hours_worked {
            label "Hour to date"
            display_template "@revisions.actual_hours_worked@ hrs"
        }
        estimated_hours_work_min {
            label "Work estimate"
            display_template "@revisions.estimated_hours_work_min@ - @revisions.estimated_hours_work_max@ hrs"
        }
        end_date {
            label "Deadline"
        }
    } \
    -orderby {
        revision_id {orderby revision_id}
        percent_complete {orderby percent_complete}
        end_date {orderby end_date}
        default_value revision_id,desc
    } \
    -orderby_name orderby_revisions \
    -sub_class {
        narrow
    } \
    -filters {
        task_revision_id {}
        orderby_dependency {}
        orderby_dependency2 {}
    } \
    -html {
        width 100%
    }


db_multirow -extend { item_url } revisions task_revisions_query {
} {
    set item_url [export_vars -base "task-one" -override {{task_revision_id $revision_id}} -exclude {revision_id} { revision_id task_id}]
}

# Dependency info ------------------------------------------------

template::list::create \
    -name dependency \
    -multirow dependency \
    -key d_task_id \
    -elements {
        dependency_type {
            label "Type"
            display_template {
                <if @dependency.dependency_type@ eq start_before_start>
                <img border="0" src="resources/start_before_start.png">
                </if>
                <if @dependency.dependency_type@ eq start_before_finish>
                <img border="0" src="resources/start_before_finish.png">
                </if>
                <if @dependency.dependency_type@ eq finish_before_start>
                <img border="0" src="resources/finish_before_start.png">
                </if>
                <if @dependency.dependency_type@ eq finish_before_finish>
                <img border="0" src="resources/finish_before_finish.png">
                </if>
            }
        }
        d_task_id {
            label "Task"
            display_col task_title
            link_url_col item_url
            link_html { title "View this task" }
        }
        percent_complete {
            label "Status"
            display_template "@dependency.percent_complete@\%"
        }
        end_date {
            label "Deadline"
        }
    } \
    -orderby {
        percent_complete {orderby percent_complete}
        end_date {orderby end_date}
    } \
    -orderby_name orderby_dependency \
    -sub_class {
        narrow
    } \
    -filters {
        task_revision_id {}
        orderby_revisions {}
        orderby_dependency2 {}
    } \
    -html {
        width 100%
    }

db_multirow -extend { item_url } dependency dependency_query {
} {
    set item_url [export_vars -base "task-one" -override {{task_id $parent_task_id}} { task_id $d_task_id }]
}

# Dependency info (dependency other task have on this task) ------

template::list::create \
    -name dependency2 \
    -multirow dependency2 \
    -key d_task_id \
    -elements {
        dependency_type {
            label "Type"
            display_template {
                <if @dependency2.dependency_type@ eq start_before_start>
                <img border="0" src="resources/start_before_start.png">
                </if>
                <if @dependency2.dependency_type@ eq start_before_finish>
                <img border="0" src="resources/start_before_finish.png">
                </if>
                <if @dependency2.dependency_type@ eq finish_before_start>
                <img border="0" src="resources/finish_before_start.png">
                </if>
                <if @dependency2.dependency_type@ eq finish_before_finish>
                <img border="0" src="resources/finish_before_finish.png">
                </if>
            }
        }
        d_task_id {
            label "Task"
            display_col task_title
            link_url_eval {task-one?task_id=$d_task_id}
            link_html { title "View this task" }
        }
        percent_complete {
            label "Status"
            display_template "@dependency2.percent_complete@\%"
        }
        end_date {
            label "Deadline"
        }
    } \
    -orderby {
        percent_complete {orderby percent_complete}
        end_date {orderby end_date}
    } \
    -orderby_name orderby_dependency2 \
    -sub_class {
        narrow
    } \
    -filters {
        task_revision_id {}
        orderby_revisions {}
        orderby_dependency {}
    } \
    -html {
        width 100%
    }


db_multirow -extend { item_url } dependency2 dependency2_query {
} {

}

# People, using list-builder ---------------------------------

db_multirow people task_people_query { }

template::list::create \
    -name people \
    -multirow people \
    -key item_id \
    -elements {
        first_names {
            label {
                Who <a href="task-assign-add-edit?revision_no_dependencies=$task_revision_id&[export_url_vars project_item_id]"><img border="0" src="/shared/images/Edit16.gif" align="center"></a>
            }
            display_template {
                @people.user_info@ 
            }
        }
        role_id {
            label "Role"
            display_template "@people.one_line@"
        }
    } \
    -sub_class {
        narrow
    } \
    -filters {
        party_id {}
        task_id {}
        orderby_subproject {}
        orderby_versions {}
        orderby_tasks {}
    } \
    -orderby {
        role_id {orderby role_id}
        default_value role_id,desc
    } \
    -orderby_name orderby_people \
    -html {
        width 100%
    }


db_multirow -extend { item_url } subproject task_people_query {
} {

}


# Xrefs ------------------------------------------------

template::list::create \
    -name xrefs \
    -multirow xrefs \
    -key x_task_id \
    -elements {
        x_task_id {
            label "ID"
        }
        title {
            label "Task"
            link_url_col item_url
            link_html { title "View this task" }
        }
        slack_time {
            label "Slack"
        }
        earliest_start_pretty {
            label "ES"
        }
        earliest_finish_pretty {
            label "EF"
        }
        latest_start_pretty {
            label "LS"
        }
        latest_finish_pretty {
            label "ES"
        }
    } \
    -sub_class {
        narrow
    } \
    -filters {
        task_revision_id {}
        orderby_revision {}
        orderby_dependency {}
        orderby_dependency2 {}
    } \
    -html {
        width 100%
    }

db_multirow -extend { item_url earliest_start_pretty earliest_finish_pretty latest_start_pretty latest_finish_pretty slack_time } xrefs xrefs_query {
} {
    set item_url [export_vars -base "task-one" -override {{task_id $x_task_id}}]

    set earliest_start_pretty [lc_time_fmt $earliest_start "%x"]
    set earliest_finish_pretty [lc_time_fmt $earliest_finish "%x"]
    set latest_start_pretty [lc_time_fmt $latest_start "%x"]
    set latest_finish_pretty [lc_time_fmt $latest_finish "%x"]

    set slack_time [pm::task::slack_time \
                        -earliest_start_j $earliest_start_j \
                        -today_j $today_j \
                        -latest_start_j $latest_start_j]

}


ad_return_template

# ------------------------- END OF FILE ------------------------- #
